// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// Om2pha.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Task to produce a multiple spectrum file containing
//              the count rates from OM filters
//
#include "Tracer.h"
#include "errstr.h"
#include "param.h"
#include "slatec.h"

#include "CalServer.h"
#include "Dal.h"


#include "Om2pha.h"


/* Method: Om2pha() - Constructor */
Om2pha::Om2pha()
{
///  PROC;

  /* Get the parameters from the command line */
    double ra = realParameter("ra");
    double dec = realParameter("dec");
    string srclist = stringParameter("srclist");
    string outputfile = stringParameter("output");

  /* Get the filter names, count rates and errors */
    FilterList filterList = getFilterRates(ra, dec, srclist);

  /* Create the output file */
    writeMultipleSpecFile(outputfile, filterList);
}

/* Destructor */
Om2pha:: ~Om2pha()
{
  // Do nothing
};
/* 
 * Method: getFilterRates() 
 * Purpose: extract filter count rates from an OM source list
*/
Om2pha::FilterList Om2pha::getFilterRates(double ra, double dec, string srclist)
{
  /* Instantiate a new filter vector */
    Om2pha::FilterList filterList;

  /* Initialise filter names and response matrices */
    initFilterNames();

  /* Attempt to open the source list */
    DataSet * srcSet = dataSetServer->open(srclist, DataSet::Read);

  /* Validate format */
    validateSrcListFormat(srcSet);

  /* Open the source list table */
    Table * srctable = srcSet->table("SRCLIST");

  /* Read the corrected celestial coordinates */
    vector<double> racorr, decorr;
    CalOalUtility::fillVector(srctable, "RA_CORR", racorr);
    CalOalUtility::fillVector(srctable, "DEC_CORR", decorr);

  /* Find the input position in the source list */
    int sindex = getIndex(ra, dec, racorr, decorr);

  /* Set RA, DEC of actual source position - to be picked up later */
    _srcRa=racorr[sindex];
    _srcDec=decorr[sindex];

    errstr<<"Using source "<<sindex+1<<" from the source list"
           << Sas::message(Msg::App, Msg::Verbose);

  /* Read the count rates for each filter */
    vector<double> filtCrate, filtError;

    for (unsigned f=0; f<_filterName.size(); ++f)
     {
       // Attempt to read the data for this column
       string filterColName = _filterName[f]+"_CORR_RATE";
       if (srctable->hasColumn(filterColName)) {
          
        // If count rate is not null then add this filter to the output vectors
          CalOalUtility::fillVector(srctable, filterColName, filtCrate);
          CalOalUtility::fillVector(srctable, filterColName+"_ERR", filtError);

        // If a count rate exists for this filter write into the output vector
          if (filtCrate[sindex] == filtCrate[sindex]) { // NULL check

        	 Om2pha::filter filter;
        	 filter.name=_filterName[f];
        	 filter.resp=_filterResp[f];
             filter.crate=filtCrate[sindex];
             filter.error=filtError[sindex];

             errstr<<"Processing filter: "<<_filterName[f]
                   << Sas::message(Msg::App, Msg::Sparse);

             filterList.push_back(filter);
          } 
       }
     }

  /* Close down */
    delete srcSet; srcSet=0;

    return filterList;
}

/* 
 * Method: writeMultipleSpecFile() 
 * Purpose: write a multiple spectrum file with the OM filter count rates
*/
void Om2pha::writeMultipleSpecFile(string outputfile,
                                      Om2pha::FilterList filterList)
{
   // Open the output file
     DataSet * outset = dataSetServer->open(outputfile, DataSet::Create);         

   // Create the spectrum extension
     Table * spectrumtable = outset->addTable("SPECTRUM", filterList.size());

   // Add columns
     Column *specnumCol = spectrumtable->addColumn("SPEC_NUM", Column::Int16,
                                     "A spectrum identifier");
     Column *chanCol = spectrumtable->addColumn("CHANNEL", Column::Int16,
                                     "Spectral channel","channel",Dal::zero);
     Column *rateCol = spectrumtable->addColumn("RATE", Column::Real32,
                                     "count rate","counts/s",Dal::zero);
     Column *ratErrCol = spectrumtable->addColumn("STAT_ERR", Column::Real32,
                                     "count rate error","counts/s",Dal::zero);

     // Find string lengths for response and RowId entries and add columns
     vector<unsigned long> rspSize, rowSize;
     rspSize.push_back(24); // Must change if response filename length increases
     rowSize.push_back(4);

     Column *respCol = spectrumtable->addColumn("RESPFILE", Dal::DString,
                                     "Response file","",rspSize);
     Column *rowidCol = spectrumtable->addColumn("ROWID", Dal::DString,
                                     "Spectrum text identifier","",rowSize);

   // Access the data
     ColumnData * specColData = specnumCol->data(),
		        *respColData = respCol->data(),
		        *rowidColData = rowidCol->data();

     int16      * specNum = specColData->int16Data(); //CHANGE

   // Fill the columns
     for (unsigned f=0; f<filterList.size(); ++f) {
    	 //cout<<"F: "<<f<<" "<<filterList[f].name<<endl;
         specNum[f]=f+1; // Spectrum number starting from 1

         // Xspec expects channel, rate and error columns with variable length
         CellData  * _cellDataC = chanCol->cellData(f);
         _cellDataC->size(1);
         _cellDataC-> int16Data()[ 0 ] = 1;
         CellData  * _cellDataR = rateCol->cellData(f);
         _cellDataR->size(1);
         _cellDataR-> real32Data()[ 0 ] = filterList[f].crate;
         CellData  * _cellDataE = ratErrCol->cellData(f);
         _cellDataE->size(1);
         _cellDataE-> real32Data()[ 0 ] = filterList[f].error;

         CalOalUtility::stringColumnRowSet(respColData, f, filterList[f].resp);
         CalOalUtility::stringColumnRowSet(rowidColData, f, filterList[f].name);
     }

   // Add keywords
     addKeys(spectrumtable);

   // Close down
     delete specColData; specColData=0;
     delete respColData; respColData=0;
     delete rowidColData; rowidColData=0;
   
     errstr << "Closing output spectrum: " << outputfile
            << Sas::message(Msg::App, Msg::Sparse);

     dataSetServer->close(outset);
}

/* 
 * Method: addKeys() 
 * Purpose: add a predefined set of keywords to the output file
*/
void Om2pha::addKeys(Table * spectrumtable)
{

    *spectrumtable->addAttribute("HDUVERS1","Obsolete - included for backwards compatibility")     = "1.2.1";
    *spectrumtable->addAttribute("HDUVERS")  = "1.2.1";
    *spectrumtable->addAttribute("CHANTYPE") = "PI";
    *spectrumtable->addAttribute("HDUCLAS1") = "SPECTRUM";
    *spectrumtable->addAttribute("HDUCLASS") = "OGIP";
    *spectrumtable->addAttribute("INSTRUME") = "OM";
    *spectrumtable->addAttribute("TELESCOP") = "XMM";
    *spectrumtable->addAttribute("HDUCLAS2") = "TOTAL";
    *spectrumtable->addAttribute("HDUCLAS3") = "RATE";
    *spectrumtable->addAttribute("HDUCLAS4") = "TYPE:II";
    *spectrumtable->addAttribute("DETCHANS") = 1;
    *spectrumtable->addAttribute("BACKSCAL") = 1.0;
    *spectrumtable->addAttribute("AREASCAL") = 1.0;
    *spectrumtable->addAttribute("POISSERR") = false;
    *spectrumtable->addAttribute("CORRSCAL") = 0.0;
    *spectrumtable->addAttribute("BACKFILE") = "NONE";
    *spectrumtable->addAttribute("ANCRFILE") = "NONE";
    *spectrumtable->addAttribute("EXPOSURE") = 1.0; // Dummy value - not used
    *spectrumtable->addAttribute("SRC_RA","RA of source (degrees)") = _srcRa;
    *spectrumtable->addAttribute("SRC_DEC","DEC of source (degrees)") = _srcDec;

}
/*
 * Method: initFilterNames()
 * Purpose: sets the names and response matrices of OM filters
*/
void Om2pha::initFilterNames()
{
	const char *filterName[] = {"UVW2","UVM2","UVW1","U","B","V"};

	const char *filterResp[] = {"om_effarea_uvw2_v2.0.rsp",
	              "om_effarea_uvm2_v2.0.rsp",
	               "om_effarea_uvw1_v2.0.rsp",
	               "om_effarea_u_v2.0.rsp",
	               "om_effarea_b_v2.0.rsp",
	               "om_effarea_v_v2.0.rsp"};

	for (unsigned f=0; f<NUM_OF(filterName); ++f) {
        _filterName.push_back(filterName[f]);
        _filterResp.push_back(filterResp[f]);
	}
}
/*
 * Method: validateSourceList()
 * Purpose: validate the source list format
*/
void Om2pha::validateSrcListFormat(DataSet * srcSet)
{
   if (srcSet->hasBlock("SRCLIST")) {

      Table * srctable = srcSet->table("SRCLIST");

    /* Test that the columns exist in the srclist table */
      if (!srctable->hasColumn("RA_CORR") || !srctable->hasColumn("DEC_CORR"))
         errstr<<"Position columns, RA_CORR or DEC_CORR not found in "
        	   <<"the source list"<< Sas::error("InvalidSrcList");

   } else {
	  errstr<<"Table SRCLIST does not exist in "
	           	   <<"the source list"<< Sas::error("InvalidSrcList");
   }
}

/*
 * Method: getIndex()
 * Purpose: find the position of this source in the source list
*/
int Om2pha::getIndex(double ra, double dec,
		                vector<double> racorr, vector<double> decorr)
{
	int index=0; // The index to be returned

	// Get the tolerence from the parameter list
	double tol = realParameter("tolerance");

	// Find sources in the list whose position is coincident with the
	// required position within the tolerance
	int nfound=0;
	double minoff=tol; // The minimum offset
	for (unsigned i=0; i<racorr.size(); ++i) {

	    // Calculate offset in arcseconds
		double offset = sqrt( (ra-racorr[i])*(ra-racorr[i])+
				              (dec-decorr[i])*(dec-decorr[i]) ) * 3600.0;
		if (offset < tol)
		{
		  // Increment counter
		   nfound++;

		  // If this is closer than the last one - use this value
		   if (offset < minoff) {
			  index = i;
			  minoff = offset;
		   }
		}
	}

	/* Has an acceptable source been found ? */
	if (nfound == 0)
		errstr<<"The source list does not contain any sources within "
		      <<tol<<" arcsecs of the position: RA="<<ra
		      <<" DEC="<<dec<< Sas::error("NoSourceFound");
	else if (nfound > 1)
		errstr<<nfound<<" entries were found in the source list within "
		      <<tol<<" arcsecs of the position: RA="<<ra
		      <<" DEC="<<dec<<" , returning the closest source - number "
		      <<index+1<< Sas::warning("MultipleSourcesFound");

	return index;
}


